/*
 * cog.h
 * Copyright (C) 2021 Igalia S.L.
 * Copyright (C) 2018 Eduardo Lima <elima@igalia.com>
 * Copyright (C) 2017-2018 Adrian Perez <aperez@igalia.com>
 *
 * Distributed under terms of the MIT license.
 */

#ifndef COG_H
#define COG_H

#define COG_INSIDE_COG__ 1

#include "cog-config.h"
#include "cog-directory-files-handler.h"
#include "cog-gamepad.h"
#include "cog-host-routes-handler.h"
#include "cog-modules.h"
#include "cog-platform.h"
#include "cog-prefix-routes-handler.h"
#include "cog-request-handler.h"
#include "cog-shell.h"
#include "cog-utils.h"
#include "cog-webkit-utils.h"

#undef COG_INSIDE_COG__

#endif /* !COG_H */
